from .. import messages as msg
from ..resources import sym
from .run import run_command


# TODO this should be re-used in changelogger prepare_new_section
def get_latest_tag(fetch: bool = True) -> str | None:
    if fetch:
        _, _ = run_command(['git', 'fetch', '--tags'], print_stderr=False)
    res, success = run_command(['git', 'describe', '--tags', '--abbrev=0'], print_stderr=False)
    if success:
        return res.stdout.strip()
    return None


def get_repo_root():
    result, success = run_command(['git', 'rev-parse', '--show-toplevel'], print_stderr=False)
    if not success:
        msg.failure('not inside a Git repository.')
    # else:
    #     print(f"{sym.positive} Inside a Git repository.")
    return result.stdout.strip()


def commit_files(config, MESSAGE):
    msg.imsg('committing file changes', icon=sym.item)

    if not config['dry_run']:
        _, success = run_command(
            ['git', 'add', 'pyproject.toml', 'uv.lock', config['changelog_path']], cwd=config['repo_root']
        )
        msg.failure('failed to add files to git') if not success else None

        _, success = run_command(['git', 'commit', '-m', MESSAGE], cwd=config['repo_root'])
        msg.failure('failed to commit changes') if not success else None


def create_git_tag(config, TAG, MESSAGE):
    msg.imsg(f'creating git tag: {TAG}', icon=sym.item)

    if not config['dry_run']:
        _, success = run_command(['git', 'tag', TAG, '-m', MESSAGE], cwd=config['repo_root'])
        msg.failure('failed to create git tag') if not success else None


def push_changes(config, TAG):
    msg.imsg('pushing to remote repository', icon=sym.item)

    if not config['dry_run']:
        _, success = run_command(['git', 'push'], cwd=config['repo_root'])
        msg.failure('failed to push file changes') if not success else None

        _, success = run_command(['git', 'push', 'origin', TAG], cwd=config['repo_root'])
        msg.failure('failed to push tag') if not success else None



from . import commands as cmd
from . import messages as msg
from .resources import ac, sym


def run_preflight(config, TAG, skip_input: bool = False):
    msg.imsg('running preflight checks...', icon=None, color=ac.BLUE)
    # check branch
    check_release_branch(config['release_branch'])

    # check tag status
    check_tags(TAG, config['repo_root'])

    # check working tree status
    check_worktree(config['repo_root'], config['allow_dirty'], skip_input=skip_input)

    # all preflight checks passed
    msg.imsg('preflight passed!\n', icon=sym.positive)

    # # show reminders if any
    # show_reminders(config['reminders'])


def check_release_branch(release_branch: str):
    if release_branch is False:
        msg.warning('skipping branch check as per configuration [release_branch = false].')
        return

    result, success = cmd.run_command(['git', 'rev-parse', '--abbrev-ref', 'HEAD'])
    if not success:
        msg.warning('failed to determine current branch!')
        return

    branch = result.stdout.strip()
    if branch != release_branch:
        msg.failure(f'you are on branch "{branch}". uv-ship config requires "{release_branch}"!')
        return
    else:
        msg.imsg(f'on release branch "{branch}".', icon=sym.positive)
        return


def check_worktree(repo_root, allow_dirty: bool = False, skip_input: bool = False):
    """Check for staged/unstaged changes before continuing."""
    result, _ = cmd.run_command(['git', 'status', '--porcelain'], cwd=repo_root)
    lines = result.stdout.splitlines()

    if not lines:
        print('✓ working tree clean.')
        tree_clean = True  # clean working tree

    else:
        proceed_dirty = False
        tree_clean = False

        staged = [line for line in lines if line[0] not in (' ', '?')]  # first column = staged
        unstaged = [line for line in lines if line[1] not in (' ', '?')]  # second column = unstaged
        untracked = [line for line in lines if line[0:2] == '??']  # second column = unstaged

        if staged:
            if not allow_dirty:
                print(f'{sym.negative} you have staged changes. Please commit or unstage them before proceeding.')
            else:
                proceed_dirty = True

        if unstaged or untracked:
            if not allow_dirty:
                confirm = (
                    'y'
                    if skip_input
                    else input(f'{sym.warning} you have unstaged changes. Proceed anyway? [y/N]: ').strip().lower()
                )
                if confirm not in ('y', 'yes'):
                    msg.abort_by_user()
                else:
                    tree_clean = True
            else:
                proceed_dirty = True

        if proceed_dirty:
            print(f'{sym.warning} proceeding with uncommitted changes. [allow_dirty = true]')
            tree_clean = True

    exit(1) if not tree_clean else None


def check_tags(tag, repo_root):
    local_result, _ = cmd.run_command(['git', 'tag', '--list', tag], cwd=repo_root)
    remote_result, _ = cmd.run_command(['git', 'ls-remote', '--tags', 'origin', tag], cwd=repo_root)

    if remote_result.stdout.strip():
        msg.failure(f'tag {tag} already exists on the remote.')

    if local_result.stdout.strip():
        confirm = (
            input(f'{sym.warning} Tag {ac.BOLD}{tag}{ac.RESET} already exists locally. Overwrite? [y/N]: ')
            .strip()
            .lower()
        )
        if confirm not in ('y', 'yes'):
            msg.abort_by_user()
            tag_clear = False

        else:
            print(f'{sym.item} deleting existing local tag {tag}')
            cmd.run_command(['git', 'tag', '-d', tag], cwd=repo_root)
            tag_clear = True
    else:
        print(f'{sym.positive} no tag conflicts.')
        tag_clear = True

    exit(1) if not tag_clear else None


def show_reminders(reminders):
    if reminders:
        print('\n', end='')
        print('you have set reminders in your config:')
        for r in reminders or []:
            print(f'{sym.item} {r}')



import importlib.resources as resources
import os
from pathlib import Path

import tomli as tomllib

from .. import messages as msg
from ..resources import sym


def _get_settings_from_toml(file: Path):
    if not file.exists():
        return None
    with open(file, 'rb') as f:
        data = tomllib.load(f)
    return data.get('tool', {}).get('uv-ship')


def load_config(path: str | None = None, cwd: str = os.getcwd(), cmd_args: dict = {}):
    """
    Load uv-ship configuration with the following precedence:
    1. Explicit path (if provided)
    2. uv-ship.toml (in cwd)
    3. pyproject.toml (in cwd, must contain [tool.uv-ship])

    Rules:
    - If both uv-ship.toml and pyproject.toml contain [tool.uv-ship], raise an error.
    - If no [tool.uv-ship] is found, prompt for a config path.
    """

    if not isinstance(cwd, Path):
        cwd = Path(cwd)

    # Find default settings and load
    def_path = resources.files('uv_ship.config')
    for cont in def_path.iterdir():
        if cont.name == 'default_config.toml':
            default_settings = _get_settings_from_toml(cont)

    # 1. If user provides a custom path → always use that
    if path:
        config_file = Path(path)

        if config_file.suffix != '.toml':
            msg.failure(f'config file "{config_file}" is not a .toml file.')

        if not config_file.exists():
            msg.failure(f'config file "{config_file}" not found.')

        settings = _get_settings_from_toml(config_file)
        if not settings:
            msg.failure(f'failed to load config! no [tool.uv-ship] table found in "{config_file}"!')

        source = config_file

    # 2. No custom path → check default files in cwd
    else:
        uv_bump_file = cwd / 'uv-ship.toml'
        pyproject_file = cwd / 'pyproject.toml'

        if not uv_bump_file.exists() and not pyproject_file.exists():
            msg.failure('could not find "uv-ship.toml" or "pyproject.toml". please provide a config path!')

        uv_bump_settings = _get_settings_from_toml(uv_bump_file)
        pyproject_settings = _get_settings_from_toml(pyproject_file)

        if uv_bump_settings and pyproject_settings:
            msg.failure(
                f'{sym.negative} Conflict: Both "uv-ship.toml" and "pyproject.toml" contain a [tool.uv-ship] table. '
                'Please remove one or specify a config path explicitly.'
            )

        if uv_bump_settings:
            settings = uv_bump_settings
            source = uv_bump_file.name

        if pyproject_settings:
            settings = pyproject_settings
            source = pyproject_file.name

        if not (uv_bump_settings or pyproject_settings):
            source = 'default'
            settings = {}

    msg.imsg(f'config source: "{source}"', color=None)
    default_settings.update(settings)

    default_settings = {k.replace('-', '_'): v for k, v in default_settings.items()}

    config = default_settings if default_settings else exit(1)

    # args = {k.replace('-', '_'): v for k, v in config.items()}
    config.update({k: v for k, v in cmd_args.items() if v is not None})

    config['repo_root'] = str(cwd)

    return config



import re
from datetime import date
from pathlib import Path

from . import commands as cmd
from . import messages as msg

HEADER_ANY = re.compile(r'^#{1,6}\s+.*$', re.M)
H_LVL = 2

tag_format = '`'


def get_commits():
    _, has_commits = cmd.run_command(['git', 'rev-parse', '--quiet', '--verify', 'HEAD'], print_stderr=False)
    if not has_commits:
        return 'there are no commits in the repository yet...'

    tag_res, has_tag = cmd.run_command(['git', 'describe', '--tags', '--abbrev=0'], print_stderr=False)
    base = tag_res.stdout.strip() if has_tag else None

    # if there is no tag, show all commits
    if base:
        log_args = ['git', 'log', f'{base}..HEAD', '--pretty=format:- %s']
    else:
        log_args = ['git', 'log', '--pretty=format:- %s']

    result, _ = cmd.run_command(log_args, print_stdout=False)
    return result.stdout


def read_changelog(config: dict, clog_path: str | Path = None, tag_format: str = tag_format) -> str:
    if not clog_path:
        clog_path = Path(config['repo_root']) / config['changelog_path']

    p = Path(clog_path) if isinstance(clog_path, str) else clog_path

    if not p.exists():
        first_section = '## beginning of changelog'  # prepare_new_section('latest', add_date=True)
        p.write_text(f'# Changelog\n\n{first_section}', encoding='utf-8')

    return p.read_text(encoding='utf-8'), clog_path


def _header_re(tag: str, level: int = H_LVL) -> re.Pattern:
    tag = f'{tag_format}{tag}{tag_format}'
    hashes = '#' * level
    # start of line, "## ", the tag, then either space/end/dash, then the rest of the line
    return re.compile(
        rf'^{re.escape(hashes)}\s+{re.escape(tag)}(?=\s|$|[-–—]).*$',
        re.M,
    )


def _normalize_bullets(text: str) -> str:
    # Ensure each non-empty line starts with "- " and trim spaces
    lines = []
    for raw in text.splitlines():
        s = raw.strip()
        if not s:
            continue
        if not s.startswith('- '):
            s = '- ' + s.lstrip('-•* ').strip()
        lines.append(s)
    return '\n'.join(lines) + '\n'


def find_section_spans(content: str, tag: str, level: int = H_LVL):
    """
    Find all sections with the given tag and level.
    Returns a list of (start, end) tuples.
    """
    spans = []
    matches = list(_header_re(tag, level).finditer(content))

    for i, m in enumerate(matches):
        start = m.start()
        nxt = HEADER_ANY.search(content, pos=m.end())
        end = nxt.start() if nxt else len(content)
        spans.append((start, end))

    return spans


def prepare_header(tag: str, add_date: bool = True, level: int = H_LVL) -> str:
    today = date.today().isoformat() if add_date else None
    header_line = f'{"#" * level} {tag_format}{tag}{tag_format}'
    if today:
        header_line += f' — [{today}]'
    return header_line


def prepare_new_section(new_tag: str, add_date: bool = True, level: int = H_LVL) -> str:
    header_line = prepare_header(new_tag, add_date=add_date, level=level)

    commits = get_commits()
    if len(commits) == 0:
        commits = '- (no changes since last tag)'

    body = _normalize_bullets(commits)
    new_section = f'{header_line}\n\n{body}\n'
    return new_section


def show_changelog(content: str, clog_file: str, print_n_sections: int | None, level: int = H_LVL):
    if print_n_sections is not None:
        # split on section headers of the same level
        section_re = re.compile(rf'^(#{{{level}}}\s+.*$)', re.M)
        parts = section_re.split(content)

        report_n = print_n_sections if print_n_sections != 1 else 'latest'
        first_line = f'\n{msg.ac.BOLD}Updated {clog_file}{msg.ac.RESET} (showing {report_n} sections)\n\n'

        rendered = [first_line]
        for i in range(1, len(parts), 2):  # step through header/body pairs
            rendered.append(parts[i])  # header
            rendered.append(parts[i + 1])  # body
            if len(rendered) // 2 >= print_n_sections:
                break
        print(''.join(rendered).strip())
    else:
        print(content)


def _insert_content(content: str, new_section: str, span: tuple[int, int]) -> str:
    return content[: span[0]] + new_section + content[span[1] :]


def get_headers(clog_content: str):
    headers = HEADER_ANY.finditer(clog_content)

    clog_headers = []
    for header in headers:
        if header.group(0).startswith('##'):
            clog_headers.append(header)

    return clog_headers


def get_latest_clog_tag(clog_content: str, tag_format: str = tag_format) -> str:
    # headers = HEADER_ANY.finditer(clog_content)
    headers = get_headers(clog_content)
    first_header = next((h for h in headers), None)
    first_clog_tag = first_header.group().removeprefix('## ').split(' — ')[0]
    return first_clog_tag.strip(tag_format)


def replace_section(clog_content: str, new_section: str, span: tuple[int, int]):
    res = _insert_content(clog_content, new_section, span)
    return res


def strategy_apply(clog_content: str, new_tag: str):
    span = get_headers(clog_content)[0].span()
    new_header = prepare_header(new_tag)
    updated = replace_section(clog_content, new_header, span)
    return updated


def strategy_update(clog_content: str, new_section: str):
    headers = get_headers(clog_content)
    first_header = next((h for h in headers), None)
    span = first_header.start(), first_header.start()

    res = _insert_content(clog_content, new_section, span)
    return res


def strategy_replace(clog_content: str, new_section: str, latest_clog_tag: str):
    spans = find_section_spans(clog_content, latest_clog_tag)
    if len(spans) > 1:
        print(f'Warning: Found multiple sections for tag {latest_clog_tag}. Replacing the first one.')
    clog_updated = replace_section(clog_content, new_section, spans[0])
    return clog_updated


def eval_clog_update_strategy(clog_content: str, new_tag: str, print_eval: bool = False):
    latest_repo_tag = cmd.git.get_latest_tag()
    latest_clog_tag = get_latest_clog_tag(clog_content)

    strategy = 'unknown'
    if not latest_repo_tag:
        if print_eval:
            print('No tags found in repository. Can not evaluate changelog update strategy.')
        strategy = 'update'
    elif latest_clog_tag == latest_repo_tag:
        if print_eval:
            print('The changelog has not been updated since the last release.')
        strategy = 'update'
    elif latest_clog_tag in ('latest', new_tag):
        if print_eval:
            print(
                'The changelog was already updated since the last release, do you want to apply this tag, or refresh it?'
            )
        strategy = 'prompt'
    else:
        msg.warning(
            f'latest changelog tag ({latest_clog_tag}) does not match latest Git tag ({latest_repo_tag}).',
        )
        strategy = 'update'
    return strategy, latest_repo_tag, latest_clog_tag


def execute_update_strategy(config, clog_path, clog_content, new_tag, strategy, save):
    new_section = prepare_new_section(new_tag, add_date=True)

    if strategy == 'prompt':
        print('It looks like the changelog was already updated since the last release (`latest` is present).')

        msg.imsg('run: `uv-ship log --latest` to compare with latest commits\n', icon=msg.sym.item, color=msg.ac.BLUE)

        if new_tag == 'latest':
            confirm = input('do you want to refresh the changelog? [y|N]:').strip().lower()

            if confirm in ('y', 'yes'):
                strategy = 'replace'

        else:
            confirm = input(f'refresh changelog or apply tag {new_tag} to that section? [r|A]:').strip().lower()
            if confirm in ('r', 'replace'):
                strategy = 'replace'
            else:
                strategy = 'apply'

    if strategy == 'update':
        clog_updated = strategy_update(clog_content, new_section)
    elif strategy == 'replace':
        clog_updated = strategy_replace(clog_content, new_section, 'latest')
    elif strategy == 'apply':
        clog_updated = strategy_apply(clog_content, new_tag)
    else:
        msg.failure(f'unknown changelog update strategy: {strategy}')

    show_changelog(content=clog_updated, clog_file=config['changelog_path'], print_n_sections=3)
    print('')

    if save:
        clog_path.write_text(clog_updated, encoding='utf-8')



import rich_click

# rich_click.rich_click.THEME = 'modern'
rich_click.rich_click.MAX_WIDTH = 85

rich_click.rich_click.COMMANDS_PANEL_TITLE = 'commands'
rich_click.rich_click.OPTIONS_PANEL_TITLE = 'options'
rich_click.rich_click.COMMANDS_BEFORE_OPTIONS = True

rich_click.rich_click.STYLE_OPTION = 'bold blue'
rich_click.rich_click.STYLE_ARGUMENT = 'bold blue'
rich_click.rich_click.STYLE_COMMAND = 'bold blue'
rich_click.rich_click.STYLE_SWITCH = 'bold red'
rich_click.rich_click.STYLE_METAVAR = 'bold red'
rich_click.rich_click.STYLE_METAVAR_SEPARATOR = 'dim'
rich_click.rich_click.STYLE_USAGE = 'bold yellow'
rich_click.rich_click.STYLE_USAGE_COMMAND = 'bold'
rich_click.rich_click.STYLE_HELPTEXT_FIRST_LINE = ''
rich_click.rich_click.STYLE_HELPTEXT = 'dim'
rich_click.rich_click.STYLE_OPTION_DEFAULT = 'dim'
rich_click.rich_click.STYLE_REQUIRED_SHORT = 'bold yellow'
rich_click.rich_click.STYLE_REQUIRED_LONG = 'bold yellow'
rich_click.rich_click.STYLE_OPTIONS_PANEL_BORDER = 'dim'
rich_click.rich_click.STYLE_COMMANDS_PANEL_BORDER = 'dim'



import subprocess


def run_command(args: list, cwd: str = None, print_stdout: bool = False, print_stderr: bool = True):
    result = subprocess.run(
        args,
        cwd=cwd,
        capture_output=True,
        text=True,
    )
    if print_stdout and result.stdout:
        print(result.stdout)
    if print_stderr and result.returncode != 0:
        # print('Exit code:', result.returncode)
        print('Error:', result.stderr)
    return result, result.returncode == 0



from . import changelogger as cl
from . import commands as cmd
from . import messages as msg
from . import preflight as prf
from .resources import ac, sym


def ship(config: dict, version: str, allow_dirty: bool = None, **kwargs):
    # dry run to collect all info first
    package_name, old_version, new_version = cmd.gen.collect_info(version=version)

    print(f'{package_name} {ac.BOLD}{ac.RED}{old_version}{ac.RESET} → {ac.BOLD}{ac.GREEN}{new_version}{ac.RESET}\n')

    # Construct tag and message
    TAG = cmd.gen.tag(config, new_version)
    MESSAGE = cmd.gen.commit_message(config, old_version, new_version)

    config['allow_dirty'] = allow_dirty if allow_dirty is not None else config['allow_dirty']

    # run preflight checks
    prf.run_preflight(config, TAG)

    confirm = input(f'{ac.BLUE}auto update changelog?{ac.RESET} [y/N]: ').strip().lower()
    if confirm in ('y', 'yes'):
        save = not config['dry_run']
        cmd_log(config=config, new_tag=TAG, save=save, show_result=1)
        print('')
        msg.imsg('please consider making manual edits NOW!', icon=sym.item, color=ac.YELLOW)
    else:
        msg.imsg('changelog update skipped by user.', icon=sym.item)

    # show preflight summary
    msg.preflight_complete()

    # Interactive confirmation
    msg.user_confirmation()

    # # TODO test safeguards
    cmd.gen.update_files(config, package_name, new_version)

    cmd.git.commit_files(config, MESSAGE)

    cmd.git.create_git_tag(config, TAG, MESSAGE)

    cmd.git.push_changes(config, TAG)

    msg.success(f'done! new version {new_version} registered and tagged.\n')


def cmd_log(config: dict, new_tag: str, latest: bool = False, save: bool = False, **kwargs):
    clog_content, clog_path = cl.read_changelog(config=config)
    strategy, latest_repo_tag, latest_clog_tag = cl.eval_clog_update_strategy(clog_content, new_tag, print_eval=False)
    # print(f'changelog update strategy: {strategy}')

    if latest:
        print('')
        msg.imsg(f'commits since last tag in repo: {latest_repo_tag}:\n', color=msg.ac.BOLD)
        new_section = cl.prepare_new_section(new_tag, add_date=True)
        print(new_section.strip())

        print('')
        msg.imsg(f'run: `uv-ship log --save` to add this to {config["changelog_path"]}\n', color=msg.ac.BLUE)

    else:
        save = save if not config['dry_run'] else False
        cl.execute_update_strategy(config, clog_path, clog_content, new_tag, strategy, save)


def cmd_status(config: dict):
    clog_content, _ = cl.read_changelog(config=config)
    latest_clog_tag = cl.get_latest_clog_tag(clog_content=clog_content)
    latest_repo_tag = cmd.git.get_latest_tag()
    repo_root = cmd.git.get_repo_root()
    project_name, current_version = cmd.gen.get_version_str(return_project_name=True)

    tool_versions = cmd.ver.get_tool_versions()

    def print_k_v(item, value):
        msg.imsg(f'{item:16}', color=msg.ac.BLUE, end='')
        msg.imsg(f'- {value}', color=msg.ac.BOLD)

    print('\nInstalled tool versions:')
    for k, v in tool_versions.items():
        if v:
            msg.imsg(f'{k:8}: {v}', color=msg.ac.DIM)

    print('')
    print_k_v('project', f'{project_name} v{current_version}')
    print_k_v('repo_root', repo_root)
    print_k_v('latest repo tag', latest_repo_tag)
    print_k_v('latest clog tag', latest_clog_tag)
    print('')



import importlib.resources as resources
import json

from .rich_click_theme import rich_click as rich_click

path = resources.files('uv_ship.resources')
for cont in path.iterdir():
    if cont.name == 'symbols.json':
        with open(cont, 'r') as f:
            symbols = json.load(f)

    if cont.name == 'ansi.json':
        with open(cont, 'r') as f:
            ansi_codes = json.load(f)

    if cont.name == 'tagline.md':
        with open(cont, 'r') as f:
            tagline = cont.read_text(encoding='utf-8')


class Symbols:
    def __init__(self, symbols):
        self.symbols = symbols
        for key, value in symbols.items():
            setattr(self, key, value)


class Ansi:
    def __init__(self, ansi_codes):
        self.ansi_codes = ansi_codes
        for key, value in ansi_codes.items():
            setattr(self, key, value)


sym = Symbols(symbols)
ac = Ansi(ansi_codes)
